package BK::Batcher;

=head1 NAME

BK::Batcher - class to support the server side of BK

=head1 SYNOPSIS

     require BK::Batcher;

=head1 DESCRIPTION

=cut

require 5.002;
use BK::Composer;
require BK::DateManager;
use POSIX;

sub new {
    my ($proto, $server) = @_;
    my $class = ref($proto) || $proto;
    my $self  = {
        SERVER      => $server,
    };
    bless ($self, $class);
    return $self;
}

sub logMesg {
    my ($self, $mesg) = @_;
    my $server = $self->{SERVER};

    if ($server->fd()) {
        print {$server->fd()} &generate({
                     data => $mesg,
                     func => 'client-OutputInfo',
           });
    } else {
	print $mesg, "\n";
    }
}

#
# Do the recurring transactions.
#
sub sortByIdx {
	my ($self, $idx1, $idx2) = @_;
	my @A = split(/\t/, $a);
	my @B = split(/\t/, $b);
	$A[$idx1] cmp $B[$idx1] || $A[$idx2] cmp $B[$idx2];
}

sub runBatch {
    my ($self, $args) = @_;
    my ($date, $mID, $precmd) = ();

    if ($args) {
	while (@$args) {
	    my $opt = shift(@$args);
	    $date = shift(@$args), next if ($opt eq '-d');
	    $mID  = shift(@$args), next if ($opt eq '-m');
            unshift(@$args, $opt);
	    last;
        }
	$precmd = join(' ', @$args) if (@$args);
    }
    my $dm = BK::DateManager->new($date);
    my $today = $dm->makeDate("%Y.%m%d");
    my $server = $self->{SERVER};

    my ($num, %indices) = (0);
    for (@{($server->getHeader('Recurrence'))[0]}) {
	$indices{$_} = $num++;
    }

    my $search = "Nextdate ne null && NextDate le $today && Interval ne never";
    $search = "Recurrence eq '$mID' && NextDate eq null && Interval eq never"
	if (defined $mID);

    my @e = sort { sortByIdx $self, $indices{'NextDate'},
			$indices{'Recurrence'} }
		$server->getEntries('Recurrence', $search);
    die "Memorized transaction \"$mID\" doesn't exist (or duplicate found)"
	if (defined $mID && @e != 1);

    while (@e) {
        my $e = $self->runEntry(shift @e, \%indices, $dm, $today, $precmd);
	if ($e) {
	    unshift(@e, $e);
	    @e = sort { sortByIdx $self,
			    $indices{'NextDate'}, $indices{'Recurrence'} } @e;
	}
    }
}

sub runEntry {
    my ($self, $e, $indices, $dm, $today, $precmd) = @_;
    my $server = $self->{SERVER};
    my @e = split /\t/, $e;

    my $rundate = $e[$indices->{'NextDate'}] || $today;
    $self->logMesg("$e[$indices->{'Recurrence'}] running on $rundate");

    $dm->assignDates($rundate);
    $self->evalEntry($dm, $e[$indices->{'TableName'}],
                          $e[$indices->{'Template'}],
                          $e[$indices->{'Commands'}],
			  $precmd);

    # Update the entry.
    my ($onceover, $nday) = (0, '');
    if ($e[$indices->{'Interval'}] eq 'never') {
	return undef;
    } elsif ($e[$indices->{'Interval'}] eq 'once') {
        $onceover = 1;
	$e[$indices->{'Interval'}] = 'never';
    } else {
        $nday = $dm->incrDate($e[$indices->{'NextDate'}], do {
			$_ = $e[$indices->{'Interval'}];
			if (/^day$/) { 1 }
			elsif (/^week$/) { 7 }
			elsif (/^month$/) { (0, 1) }
			elsif (/^year$/) { (0, 0, 1) }
			else { die "Bad interval \"$_\"" }
		    }
    		);
    }
    $e[$indices->{'NextDate'}] = $nday;
    $e = join "\t", @e;
    if ($nday gt $today || $onceover) {
        $server->reallyDelete('Recurrence', $e[$indices->{'Recurrence'}]);
        
        my $db = $server->home() . "/Recurrence.rdb";
        # Untaint $db.
        $db =~ /(.*)/; $db = $1;

        open(DATABASE, ">>$db") || die "Can't open $db: $!\n";
        print DATABASE $e, "\n";
        close DATABASE;

	$e = undef;
        $self->logMesg($onceover
                            ? "$e[$indices->{'Recurrence'}] has been "
                              . "converted to a memorized transaction"
                            : "$e[$indices->{'Recurrence'}] will run again"
                              . " on $nday");
    }
    $e;
}

sub evalEntry {
    my ($self, $dm, $name, $tID, $cmds, $precmd) = @_;
    my $server = $self->{SERVER};

    die "Invalid table name \"$name\""
        if ($name !~ /^[A-Z][a-z]*$/);

    my ($f, $t, $d) = $server->getHeader($name);

    @e = join("\t", ('new', ('') x (@$f -1)));
    @e = $server->getEntries($name, "$name eq '$tID'")
        if (length $tID);

    die "${name}.rdb inconsisent: duplicate ID \"$tID\""
                if (@e > 1);

    @DATA = split("\t", $e[0]);
    $cmds =~ s/\\T/\t/g; $cmds =~ s/\\N/\n/g;

    $precmd = '' if not defined $precmd;
    my @cmds = split(' ', $precmd . ' ; ' . $cmds);
    foreach $arg (@cmds) {
        my $num = 0;
        foreach $field (@$f) {
            if ($arg eq $field) {
                $arg = '$DATA[' . $num . ']';
                last;
            } elsif ($arg =~ /((make|incr)Date\(?)/) {
                $arg =~ s/((make|incr)Date\(?)/\$dm->$1/g;
                last;
            }
            $num++;
        }
    }

    # Untaint $newcmd.
    my $newcmd = join(' ', @cmds);
    $newcmd =~ /(.*)/; $newcmd = $1;

    eval($newcmd);
    die $@ if $@;

    $num = 0;
    while (@$f && @$t && @$d) {
        my ($field, $type, $desc) = (shift @$f, shift @$t, shift @$d);
        $server->verify($DATA[$num++] || '', $field, $type, $desc);
    }

    my $state = 'created';
    if ($DATA[0] eq $tID || $DATA[0] eq 'new') {
        $DATA[0] = $server->newID($name);
    } else {
        $server->reallyDelete($name, $DATA[0]);
        $state = 'saved';
    }

    my $entry = join("\t", @DATA);    

    my $db = $server->home() . "/" . $name . ".rdb";
    # Untaint $db.
    $db =~ /(.*)/; $db = $1;

    open(DATABASE, ">>$db") || die "Can't open $db: $!\n";
    print DATABASE $entry, "\n";
    close DATABASE;

    # print "$name '$entry{$name}' saved.\n";
    $self->logMesg("$name '$DATA[0]' $state.");
}

1;
