package BK::ServFunc::Statement;

require 5.002;
use BK::Composer;
require BK::Document;
use POSIX;

sub register {
    my $cmd = 'statement';
    return {
	"user-$cmd"	=> { func      => \&BK::ServFunc::Statement::genDoc,
			     shorthelp => "Generate a financial statement",
			     help      => "
    $cmd -- Generate a financial statement

    bk $cmd [ -d doctype ] [ -p project ] [ -f from ] [ -t to ] stmt_type

        \u$cmd generates a financial statement file of the type stmt_type,
        containing information from the relevant Invoice, Charge and
        Expense information.  Currently, the only statement type is
        'income'.

        Use the -d option to specify a documant type.  Currently,
        the 'postscript' and 'text' formats are available.  Postscript is
        the default.

        Use the -f and -t option to specify (inclusive) dates for the
        fiscal period.  Use the -p option to get information on a
        project.
",
			   },
    };
}

sub genDoc {
    my ($server, $table, $args) = @_;
    my ($docType, $from, $to, $project, $statment) = ('postscript');

    while (1) {
	$_ = shift(@$args);
	/-d/ && do { $docType = shift(@$args); next; };
	/-f/ && do { $from    = shift(@$args); next; };
	/-p/ && do { $project = shift(@$args); next; };
	/-t/ && do { $to      = shift(@$args); next; };
        $statement = $_;
	last;
    }
    if (scalar @$args || not defined $statement) {
	die "Incorrect usage.  Try 'bk help statement'.";
    }
    $server->verify($from, "From", "9N", "@@ YYYY.MMDD")
	if defined $from;
    $server->verify($to,   "To",   "9N", "@@ YYYY.MMDD")
	if defined $to;
    if (defined $from && defined $to) {
	if ($to < $from) { ($from, $to) = ($to, $from); }
    }
    
    my $doc = BK::Document->new($server->libdir() . "/templates",
				    "\u\L$statement",
				    $docType);
    if ($statement =~ /income/i) {
	&BK::ServFunc::Statement::genIncome($server,
					    $doc, $project, $from, $to);
    } else {
	die "Unknown statement type \"$statement\"";
    }
    
    print {$server->fd()} &generate({
                     data => $doc->generate(),
                     func => 'client-OutputData',
               });
    $server->release();
}

sub genIncome {
    my ($server, $doc, $project, $from, $to) = @_;

    $doc->addInfo('Project', $project)
	if defined $project;
    $doc->addInfo('From', $from)
	if defined $from;
    $doc->addInfo('To', $to)
	if defined $to;
    $doc->itemFields('ItemName', 'ItemValue');

    my $revenue = &BK::ServFunc::Statement::doRevenue($server,
					    $doc, $project, $from, $to);
    my $expense = &BK::ServFunc::Statement::doExpense($server,
					    $doc, $project, $from, $to);
    $doc->addItem(join("\t", "INCOME", $revenue -$expense));
}

sub doRevenue {
    my ($server, $doc, $project, $from, $to) = @_;
    my $search = "Date ne null";

    $search .= " && Date >= $from"
	if defined $from;
    $search .= " && Date <= $to"
	if defined $to;

    my @invoices = $server->getEntries('Invoice', $search, 'Invoice');

    $search = 'Invoice eq \'' . join('\' || Invoice eq \'', @invoices) . '\'';
    $search = "Project eq '$project' && ( " . $search . " )"
	if defined $project;
    my @charges = $server->getEntries('Charge', $search,
				      'Quantity UnitCost Tax');

    my ($subtotal, $taxtotal, %taxtotal) = (0, 0, ());
    foreach (@charges) {
	my ($quan, $cost, $tax) = split("\t");
        $cost *= $quan;
        $subtotal += $cost;
        my %taxes = $server->calcTaxes($cost, $tax);
	while (($key,$val) = each %taxes) {
            $taxtotal{$key} += $val;
            $taxtotal += $val;
        }
    }
    $doc->addItem(join("\t", "Project Revenue", $subtotal));
    while (($key,$val) = each %taxtotal) {
	$doc->addItem(join("\t", "$key Revenue", $val));
    }
    $doc->addItem(join("\t", "TOTAL REVENUE", $subtotal +$taxtotal));

    $subtotal +$taxtotal;
}

sub doExpense {
    my ($server, $doc, $project, $from, $to) = @_;
    my $search = "Date ne null";

    $search .= " && Date >= $from"
	if defined $from;
    $search .= " && Date <= $to"
	if defined $to;
    $search .= " && Project eq '$project'"
	if defined $project;

    my @expense = $server->getEntries('Expense', $search,
				      'Expenditure BaseCost Tax');
    my %subtotal = ();
    foreach (@expense) {
	my ($type, $cost, $tax) = split("\t");
        $subtotal{$type} += $cost;
        my %taxes = $server->calcTaxes($cost, $tax);
	while (($key,$val) = each %taxes) {
            $subtotal{$type} += $val
		if $key ne 'GST';
        }
    }
    my $total = 0;
    while (($key,$val) = each %subtotal) {
	$doc->addItem(join("\t", $key, $val));
	$total += $val;
    }
    $doc->addItem(join("\t", "TOTAL EXPENSE", $total));

    $total;
}

1;
